from ase.calculators.calculator import FileIOCalculator, CalculatorSetupError
from ase.calculators.calculator import InputError
import numpy as np
from ase.units import Ang, Bohr, Ha
from os import path, rename
from functools import reduce
from qmcblip.champio import Settings


class CHAMP(FileIOCalculator):
    """
    An ASE calculator for the Quantum Monte Carlo software CHAMP
    """ 

    # Stress only for FLARE compatibility
    implemented_properties = ['energy', 'forces', 'stress']

    name = "CHAMP"

    # The parameters to run CHAMP
    default_parameters = dict(
        vmc_in='vmc.inp',
        vmc_out='vmc.out',
        force_file='write_forces',
        pos_file='molecule.xyz',
        champ_loc='/usr/bin/vmc.mov1',
        nodefile=None,
        ncore=1,
        settings=None,
        use_opt_wf=False)

    # Placeholder
    command = ""

    def __init__(self, restart=None,
                ignore_bad_restart_file=FileIOCalculator._deprecated,
                label='CHAMP', atoms=None, **kwargs):
        """
        Keyword arguments:
        vmc_in -- The input file for CHAMP (default 'vmc.inp')
        vmc_out -- The output file of CHAMP (default 'vmc.out')
        force_file -- The file that CHAMP writes the forces and energies to (default 'write_forces')
        pos_file -- The file from which CHAMP will read to location of the atoms (default 'molecule.xyz')
        champ_loc -- Location of the CHAMP executable (default '/usr/bin/vmc.mov1')
        nodefile -- If set, the calculator will run on multiple nodes for CHAMP
        ncore -- Amount of cores to run CHAMP on (default 1)
        settings -- Input settings for CHAMP (OVERRULES VMC.INP)
        use_opt_wf -- Use the optimized WF from last step (default False)
        """
        
        FileIOCalculator.__init__(self, restart, ignore_bad_restart_file,
                                label, atoms, **kwargs)

        if (not path.exists(self.parameters['champ_loc'])):
            raise CalculatorSetupError("Did not find champ at: " + self.parameters['champ_loc'])
            
        if (not path.exists(self.parameters['vmc_in']) and self.parameters['settings'] is None):
            raise CalculatorSetupError("You did no supply any configuration parameters for CHAMP. \
                                       Either give a .inp file or give a configuration dictionary.")
        
        if not isinstance(self.parameters['settings'], Settings) and self.parameters['settings'] is not None:
            raise CalculatorSetupError('You did not provide a proper settings object.')

        if self.parameters['settings'] is not None:
           self.parameters['settings'].write(filename=self.parameters['vmc_in'])
        else:
            self.parameters['settings'] = Settings.read(self.parameters['vmc_in'])

        # We use a temporary input file
        self.parameters['settings'].write('vmc_temp.inp')

        # Set the command to call CHAMP
        self._set_command()
    
    def _set_command(self):
        # Add nodefile or not
        if (self.parameters['nodefile'] is None):
            self.command = "mpirun -n " + str(self.parameters['ncore']) + " " \
                    + self.parameters['champ_loc'] + " -i vmc_temp.inp" \
                    + " -o " + self.parameters['vmc_out']
        else:
            self.command = "mpirun -s all -n " + str(self.parameters['ncore']) \
                    + " -machinefile " + self.parameters['nodefile'] + " " \
                    + self.parameters['champ_loc'] + " -i vmc_temp.inp" \
                    + " -o " + self.parameters['vmc_out']

    def configure(self, **kwargs):
        """
        (Re)configure the CHAMP calculator by setting the keyword arguments.
        The tags for the vmc.inp can also be set here.
        """
        self.set(**kwargs)

        # Set the command to call CHAMP. May have changed due to the parameters.
        self._set_command()

    def write_input(self, atoms, properties=None, system_changes=None):
        FileIOCalculator.write_input(self, atoms, properties, system_changes)

        # Write the coordinates of the atoms in xyz format
        with open(self.parameters['pos_file'], 'w') as fileobj:
            comment = atoms.symbols
            natoms = len(atoms)
            fmt='%22.15f'

            fileobj.write('%d\n%s\n' % (natoms, comment))

            for s, (x, y, z) in zip(atoms.symbols, atoms.positions):
                fileobj.write('%-2s %s %s %s\n' % (s, fmt % (x * Ang/Bohr), \
                                fmt % (y * Ang/Bohr), fmt % (z * Ang/Bohr)))

        if self.parameters['use_opt_wf']:
            self.parameters['settings'].use_opt_wf()

        # Rewrite the input file
        self.parameters['settings'].write('vmc_temp.inp')



    def read(self, label):
        raise NotImplementedError

    def read_results(self):
        if (not path.exists(self.parameters['force_file'])):
            raise InputError("Could not detect the forces from CHAMP in " \
                                        + self.parameters['force_file'])

        with open(self.parameters['force_file'], 'r') as fileobj:
            # Read the energy and convert to eV
            self.results['energy'] = float(fileobj.readline()) * Ha

            # Read the forces
            forces = []
            for line in fileobj.readlines():
                f_list = [float(i) for i in line.split(" ") if i.strip()]
                forces.append(f_list)

            # Convert to eV/A and use ASE force direction
            self.results['forces'] = -np.asarray(forces) * Ha/Bohr   
        
        # Pass an null stress tensor for FLARE compatibility
        self.results['stress'] = np.zeros((6,)) 



from os.path import exists
from os import remove, rename
from glob import glob
from pydantic import BaseModel, Field, DirectoryPath, FilePath
from typing import Optional, Type, Union
from pathlib import PosixPath, Path


class Settings(BaseModel):
    class General(BaseModel):
        title: str
        pool: DirectoryPath = Field(PosixPath('./pool/'), postfix="/")
        basis: str
        pseudopot: Optional[str]
        mode: str = 'vmc_one_mpi1'
        seed: int = 1837465927472523
        eunit: str = "Hartrees"

    class Ase(BaseModel):
        iase: int = 1
        iforce_analy: int = 0
        node_cutoff: int = 1
        enode_cutoff: float = 0.05

    class Electrons(BaseModel):
        nup: int
        nelec: int

    class Optwf(BaseModel):
        ioptwf: int = 1
        ioptci: int = 1
        ioptjas: int = 1
        ioptorb: int = 1
        nextorb: int = 100
        no_active: int = 1
        nopt_iter: int = 1
        isample_cmat: int = 0
        energy_tol: float = 0.0

    class BlockingVmc(BaseModel):
        vmc_nstep: int = 20
        vmc_nblk: int = 400
        vmc_nblkeq: int = 1
        vmc_nconf_new: int = 0

    class Pseudo(BaseModel):
        nloc: int = 4
        nquad: int = 6

    general: General
    molecule: Path = Field(prefix="load ")
    basis_num_info: FilePath = Field(prefix="load ")
    determinants: FilePath = Field(prefix="load ")
    orbitals: FilePath = Field(prefix="load ")
    jastrow: FilePath = Field(prefix="load ")
    jastrow_der: FilePath = Field(prefix="load ")
    symmetry: Optional[FilePath] = Field(prefix="load ")
    ase: Optional[Ase]
    electrons: Electrons
    optwf: Optional[Optwf]
    pseudo: Optional[Pseudo]
    blocking_vmc: Optional[BlockingVmc]

    def write(self, filename='vmc.inp'):
        """
        Write a this dataclass containing the CHAMP configuration to an input file.

        Arguments:
        filename -- input file to write to
        """
        f = open(filename, 'w')
        
        schema = self.schema()['properties']
        for ind, item in enumerate(self):
            if isinstance(item[1], BaseModel):
                schema2 = item[1].schema()['properties']
                f.write("\n%module " + item[0] + "\n")
                for ind2, item2 in enumerate(item[1]):
                    if 'postfix' in schema2[item2[0]]:
                        f.write("\t" + item2[0] + " " + str(item2[1]) + schema2[item2[0]]['postfix'] + "\n")
                    else:
                        f.write("\t" + item2[0] + " " + str(item2[1]) + "\n")
                f.write("%endmodule\n\n")
            else:
                if 'prefix' in schema[item[0]]:
                    f.write(schema[item[0]]['prefix'] + item[0] + " " + str(item[1]) + '\n')
                else:
                    f.write(item[0] + " " + str(item[1]) + '\n')

        f.close()

    @classmethod
    def read(cls: Type['Model'], filename: Union[str, Path]) -> 'Model':
        """
        Read the CHAMP input file and convert it to a dictionary format

        Arguments:
        filename -- file of the input file to read and write to

        Output:
        Settings object containing the CHAMP configuration
        """
        # Check if the file exists
        if not exists(filename):
            raise FileNotFoundError(filename + " was not found!")

        output = dict()

        f = open(filename, 'r')

        # Set some variables to keep track of modules
        curmod = ""
        inmod = False

        for line in f:
            # We skip empty lines
            if not line.strip():
                continue
            # Beginning of a module
            elif line.startswith("%module"):
                curmod = line[8:-1]
                inmod = True
                output[curmod] = dict()
            # End of a module
            elif line.startswith("%endmodule"):
                inmod = False
            # Loading in a file
            elif line.startswith("load"):
                temp = line[5:-1]
                temp = temp.split()
                key = temp[0]
                temp.pop(0)
                value = ' '.join(temp)
                output[key] = value
            # All other tags
            else:
                temp = line[:-1].split()
                key = temp[0]
                temp.pop(0)
                value = ' '.join(temp)
                if inmod:
                    # If we are in a module, add the tag to that subdictionary.
                    output[curmod][key] = value
                else:
                    output[key] = value

        f.close()

        return cls(**output)

   
    def use_opt_wf(self):
        """
        Function to replace the orbitals, determinants and jastrow with the optimized files.

        Arguments:
        filename -- file of the input file to  write to
        """
        opt = ["det_optimal.1.iter*", "orbitals_optimal.1.iter*", "jastrow_optimal.1.iter*"]
        keys = ["determinants", "orbitals", "jastrow"]

        for ind, name in enumerate(opt):
            num = len(glob(name))
            # If there are no optimized WF files, we do not use them
            index = min(num, self.optwf.nopt_iter)
            if num >= index and index > 0:
                opt[ind] = opt[ind].strip('*') + str(index)
                setattr(self, keys[ind], opt[ind])

    def todict(self):
        return self.json(exclude_none=True)

def cleanup(*args):
    """
    Remove files created by CHAMP. Add files as arguments to include (if not in default list)
    or exclude (if in default list) files.
    """
    # Default list of the files that will be removed
    standard = ["force_analytic", "restart_vmc", "output.log", "parser.log", "orbitals_optimal.1.*",
                "jastrow_optimal.1.*", "det_optimal.1.*", "geo_optimal.*", "geo_optimal_final",
                "mc_configs_start", "nohup.out", "vmc_temp.inp"]
    
    # Add or remove file depending on the input arguments
    for item in args:
        if item in standard:
            standard.remove(item)
        else:
            standard.append(item)

    for file in standard:
        # Take care of the wildcard files
        if (file.find('*') != -1):
            for f in glob(file):
                remove(f)
        else:
            if exists(file):
                remove(file)




# MIT License

# Copyright (c) 2019 Harvard University

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE

from time import time
from copy import deepcopy
import numpy as np

from flare.otf import OTF
from flare.ase.otf import ASE_OTF
from flare.ase.atoms import FLARE_Atoms
import flare.ase.dft as dft_source
from flare.utils.learner import is_std_in_bound

from ase import units

from qmcblip.verlet import CustomVerlet


class C_OTF(OTF):

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

    def run(self):
        """
        Performs an on-the-fly training run.

        If OTF has store_dft_output set, then the specified DFT files will
        be copied with the current date and time prepended in the format
        'Year.Month.Day:Hour:Minute:Second:'.
        """

        optional_dict = {"Restart": self.curr_step}
        self.output.write_header(
            str(self.gp),
            self.dt,
            self.number_of_steps,
            self.structure,
            self.std_tolerance,
            optional_dict,
        )

        counter = 0
        self.start_time = time()

        while self.curr_step < self.number_of_steps:
            # run DFT and train initial model if first step and DFT is on
            if (
                (self.curr_step == 0)
                and (self.std_tolerance != 0)
                and (len(self.gp.training_data) == 0)
            ):

                # Are the recorded forces from the GP or DFT in ASE OTF?
                # When DFT is called, ASE energy, forces, and stresses should
                # get updated.
                self.initialize_train()

            # after step 1, try predicting with GP model
            else:
                # compute forces and stds with GP
                self.dft_step = False
                self.compute_properties()

                # get max uncertainty atoms
                std_in_bound, target_atoms = is_std_in_bound(
                    self.std_tolerance,
                    self.gp.force_noise,
                    self.structure,
                    max_atoms_added=self.max_atoms_added,
                    update_style=self.update_style,
                    update_threshold=self.update_threshold,
                )

                steps_since_dft = self.curr_step - self.last_dft_step
                if (not std_in_bound) and (steps_since_dft > self.min_steps_with_model):
                    # record GP forces
                    self.update_temperature()
                    self.record_state()
                    gp_energy = self.structure.potential_energy
                    gp_forces = deepcopy(self.structure.forces)
                    gp_stress = deepcopy(self.structure.stress)

                    # run DFT and record forces
                    self.dft_step = True
                    self.last_dft_step = self.curr_step
                    self.run_dft()
                    dft_frcs = deepcopy(self.structure.forces)
                    dft_stress = deepcopy(self.structure.stress)
                    dft_energy = self.structure.potential_energy

                    # run MD step & record the state
                    self.record_state()

                    # record DFT data into an .xyz file with filename self.dft_xyz.
                    # the file includes the structure, e/f/s labels and atomic
                    # indices of environments added to gp
                    self.record_dft_data(self.structure, target_atoms)

                    # compute mae and write to output
                    self.compute_mae(
                        gp_energy,
                        gp_forces,
                        gp_stress,
                        dft_energy,
                        dft_frcs,
                        dft_stress,
                    )

                    # add max uncertainty atoms to training set
                    self.update_gp(
                        target_atoms,
                        dft_frcs,
                        dft_stress=dft_stress,
                        dft_energy=dft_energy,
                    )

                    if self.write_model == 4:
                        self.checkpoint()
                        self.backup_checkpoint()

            # write gp forces
            if counter >= self.skip and not self.dft_step:
                self.update_temperature()
                self.record_state()
                counter = 0

            counter += 1
            # TODO: Reinstate velocity rescaling.
            self.md_step(forces=self.structure.forces)  # update positions by Verlet
            self.rescale_temperature(self.structure.positions)

            if self.write_model == 3:
                self.checkpoint()

        self.output.conclude_run()

        if self.write_model >= 1:
            self.write_gp()
            self.checkpoint()

    def initialize_train(self):
        # call dft and update positions
        self.run_dft()
        dft_frcs = deepcopy(self.structure.forces)
        dft_stress = deepcopy(self.structure.stress)
        dft_energy = self.structure.potential_energy

        self.update_temperature()
        self.record_state()
        self.record_dft_data(self.structure, self.init_atoms)

        # make initial gp model and predict forces
        self.update_gp(
            self.init_atoms, dft_frcs, dft_stress=dft_stress, dft_energy=dft_energy
        )
        self.structure.forces = dft_frcs

class C_ASE_OTF(ASE_OTF, C_OTF):

    def __init__(
        self,
        atoms,
        timestep,
        number_of_steps,
        dft_calc,
        md_engine,
        md_kwargs,
        update_settings,
        calculator=None,
        trajectory=None,
        **otf_kwargs
    ):

        self.structure = FLARE_Atoms.from_ase_atoms(atoms)
        if calculator is not None:
            self.structure.calc = calculator
        self.timestep = timestep
        self.md_engine = md_engine
        self.md_kwargs = md_kwargs
        self._kernels = None
        self.update_settings = np.array(update_settings)

        if md_engine == "CustomVerlet":
            MD = CustomVerlet
        else:
            raise NotImplementedError(md_engine + " is not implemented in ASE")

        self.md = MD(
            atoms=self.structure,
            timestep=timestep,
            trajectory=trajectory,
            **md_kwargs,
        )

        force_source = dft_source
        self.flare_calc = self.structure.calc

        # Convert ASE timestep to ps for the output file.
        flare_dt = timestep / (units.fs * 1e3)

        C_OTF.__init__(
            self,
            dt=flare_dt,
            number_of_steps=number_of_steps,
            gp=self.flare_calc.gp_model,
            force_source=force_source,
            dft_loc=dft_calc,
            dft_input=self.structure,
            **otf_kwargs,
        )

        self.flare_name = self.output_name + "_flare.json"
        self.dft_name = self.output_name + "_dft.pickle"
        self.structure_name = self.output_name + "_atoms.json"
        self.checkpt_files = [
            self.checkpt_name,
            self.flare_name,
            self.dft_name,
            self.structure_name,
            self.dft_xyz,
        ]

    def md_step(self, forces=None):
        """
        Get new position in molecular dynamics based on the forces predicted by
        FLARE_Calculator or DFT calculator
        """

        if self.curr_step in self.update_settings:
            sets = self.dft_loc.parameters.settings
            ind = np.where(self.update_settings == self.curr_step)[0][0]
            changes = self.update_settings[ind][1]
            for key, val in changes.items():
                if isinstance(val, dict):
                    for key2, val2 in val.items():
                        setattr(getattr(sets, key), key2, val2)
                else:
                    setattr(sets, key, val)

        # Update previous positions.
        self.structure.prev_positions = np.copy(self.structure.positions)

        # Reset FLARE calculator.
        if self.dft_step:
            self.flare_calc.reset()
            self.structure.calc = self.flare_calc

        # Take MD step.
        # Inside the step() function, get_forces() is called
        self.md.step(forces=forces)
        self.curr_step += 1